/**
    uBlock Origin - a browser extension to block requests.
    Copyright (C) 2014-present Raymond Hill

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see {http://www.gnu.org/licenses/}.

    Home: https://github.com/gorhill/uBlock
*/

body {
    align-items: center;
    border: 1px solid var(--ubo-red);
    box-sizing: border-box;
    display: flex;
    flex-direction: column;
    height: 100vh;
    padding: 0 2px;
    position: relative;
    width: 100vw;
    }

.logo {
    left: 0;
    padding: 2px 1px;
    position: absolute;
    top: 0;
    }

#frameURL {
    font-family: monospace;
    font-size: 90%;
    overflow-y: auto;
    word-break: break-all;
    }
#frameURL > a {
    font-size: 1rem;
    }
    
#clickToLoad {
    cursor: default;
    margin-bottom: 1em;
    }
